#include "stacktool.h"

static const bool debug_avr = false;

#if USE_BDD
#include "bdd_helpers.h"
#define STACKTOOL 1
#include "bdd_ops/all_bdd_bitbit_incs.h" 
#else
#include "avr_manual.h"
#endif

/////////////////// core operations on AVR machine state ///////////////////

static void do_avr_in (memory *m,
		       const struct decoder_info *dec)
{
  reg_t reg = dec->dst_reg;
  int mask;

  assert (dec->dst_reg_set);

  /*
   * construct a mask of bits from this io address that are modeled;
   * other bits should be turned to bottom
   */
  if (dec->ioaddr == SPL_ADDR ||
      dec->ioaddr == SPH_ADDR ||
      dec->ioaddr == SREG_ADDR) {
    // model all bits of these
    mask = 0xff;
  } else {
    mask = 0x0;
    for (unsigned int i=0; i<platform->num_vectors; i++) {
      if (platform->int_vecs[i].enable == NO_BIT) continue;
      if (platform->int_vecs[i].enable.addr == io_to_mem (dec->ioaddr)) {
	mask |= 1<<(platform->int_vecs[i].enable.bit);
      }
    }
  }

  copy_mem_byte (m, 
		 io_to_mem (dec->ioaddr),
		 reg_to_mem (reg));

  m->core[reg_to_mem(reg)].known &= mask;
  m->core[reg_to_mem(reg)].val &= mask;
}

static void do_avr_movw (memory *m,
			 const struct decoder_info *dec)
{
  assert (dec->src_reg_set);
  assert (dec->dst_reg_set);
  copy_mem_byte (m, 
		 reg_to_mem (dec->src_reg), 
		 reg_to_mem (dec->dst_reg));
  copy_mem_byte (m, 
		 reg_to_mem (dec->src_reg + (reg_t)1), 
		 reg_to_mem (dec->dst_reg + (reg_t)1));
}

// We can't handle interrupts that enable other interrupts
// and then don't disable them again before they exit.
// For example, an irq handler might want to set a timer which
// might require it to enable timer interrupts.
//
// Since we can't handle it, we try to detect when it happens so
// we can warn the user.  We can't do a perfect job but the following
// works tolerably well:
//
// 1) [First attempt] In reti instructions, we check whether any
//    of the individual irqs are enabled.  (They would have been
//    bottom on entry to the handler.)
//    But, simply checking at the end isn't enough
//    because it will not detect the case that one path sets the flag
//    while another path leaves it alone.  
//
// 2) [Final version] To handle this, we add a
//    'has been set to 1' flag to each irq enable flag ('1' is the bottom
//    value for this domain - we want to know if any path sets the flag).

// implicitly initialized to all false...
bool indirectly_enabled[MAX_VECTORS];

static void check_irqs_enabled(code_addr_t PC, memory *m) {
  for (unsigned int vec=1; vec < platform->num_vectors; vec++) {
    data_addr_t a = platform->int_vecs[vec].enable.addr;
    bit_t b = platform->int_vecs[vec].enable.bit;
#if 1
    if (is_bit_set(m->core[a].known,b)
        && is_bit_set(m->core[a].val,b)) {
#else
    // Hmmm, this code doesn't work - not sure why not
    if (is_bit_set(m->modified[a],b)) {
#endif
      cout << "Warning: IRQ " << vec << " enabled at end of function "
           << addr_to_head_label (PC)
           << ". IPG might be inaccurate."
        //           << " " << hex << PC << " " << a << "/" << b 
        //           << " " << m->core[a]
           << endl;
      indirectly_enabled[vec] = true;
    }
  }
  //  cout << "Warning: finished checking for irqs that enable irqs" << endl;
}

static void do_avr_out (memory *m,
			const struct decoder_info *dec)
{
  // redundant check that's nice since it gives a better error message
  if (dec->ioaddr == SPL_ADDR || dec->ioaddr == SPH_ADDR) {
    if (m->core[reg_to_mem(dec->src_reg)].known != 0xff) {
      printf ("oops: indeterminate value written to SP at %x\n",
              dec->addr);
      die ();
    }
  }

  copy_mem_byte (m, 
		 reg_to_mem (dec->src_reg), 
		 io_to_mem (dec->ioaddr));

  mem_loc v = m->core[reg_to_mem(dec->src_reg)];

  // We keep track of any bits that have ever been set to 1
  // This is used to help us detect irq handlers that enable
  // other irq handlers.
  m->modified[io_to_mem (dec->ioaddr)] |= (v.known & v.val);
}

static void add_mem_pair (memory *m,
			  data_addr_t lo,
			  unsigned int value)
{
  assert (value <= 255);
  data_addr_t hi = lo+1;
#if USE_BDD
  const struct mem_loc_t one = mem_loc(value);
  const struct mem_loc_t zero = mem_loc(0);
  struct mem_loc_t cc = mem_loc(0);
  struct mem_loc_t new_lo = bdd_bitbit_add_res (m->core[lo],
					        one,
						cc);  
  cc = bdd_bitbit_add_cc (m->core[lo],
			  one,
			  cc);
  struct mem_loc_t new_hi = bdd_bitbit_adc_res (m->core[hi],
					        zero,
						cc);  
  m->core[lo] = new_lo;
  m->core[hi] = new_hi;
#else
  // if we care about stacktool classic, this can be improved
  kill_mem (m, lo);
  kill_mem (m, hi);
#endif
}

static void add_reg_pair (memory *m,
			  reg_t lo, 
			  unsigned int value)
{
  add_mem_pair (m, reg_to_mem (lo), value);
}

static void inc_reg_pair (memory *m,
			  reg_t lo)
{
  add_reg_pair (m, lo, 1);
}

static void inc_mem_pair (memory *m,
			  data_addr_t lo)
{
  add_mem_pair (m, lo, 1);
}

static void sub_mem_pair (memory *m,
			  data_addr_t lo,
			  unsigned int value)
{
  assert (value <= 255);
  data_addr_t hi = lo+1;
#if USE_BDD
  const struct mem_loc_t zero = mem_loc(0);
  const struct mem_loc_t one = mem_loc(value);
  struct mem_loc_t cc = mem_loc(0);
  struct mem_loc_t new_lo = bdd_bitbit_sub_res (m->core[lo],
					        one,
						cc);  
  cc = bdd_bitbit_sub_cc (m->core[lo],
			 one,
			 cc);
  struct mem_loc_t new_hi = bdd_bitbit_sbc_res (m->core[hi],
					        zero,
						cc);  
  m->core[lo] = new_lo;
  m->core[hi] = new_hi;
#else
  // if we care about stacktool classic, this can be improved
  kill_mem (m, lo);
  kill_mem (m, hi);
#endif
}

static void sub_reg_pair (memory *m,
			  reg_t lo,
			  unsigned int value)
{
  sub_mem_pair (m, reg_to_mem (lo), value);
}

static void dec_reg_pair (memory *m,
			  reg_t lo)
{
  sub_reg_pair (m, lo, 1);
}

static void dec_mem_pair (memory *m, 
			  data_addr_t lo)
{
  sub_mem_pair (m, lo, 1);
}

static void arith_16 (memory *m,
		      const struct decoder_info *dec)
{
  reg_t dst_reg = dec->dst_reg;
  assert (dec->K_set);
  assert (dec->dst_reg_set);
  assert (dst_reg == (reg_t)24 || 
	  dst_reg == (reg_t)26 ||
	  dst_reg == (reg_t)28 || 
	  dst_reg == (reg_t)30);

  int orig_high = m->core[reg_to_mem(dst_reg+1)].val;
  int orig_high_known = m->core[reg_to_mem(dst_reg+1)].known;
  
  if (dec->opcode == AVR_adiw) {

#if 0
    cout << "adiw: adding " << dec->K << " to r" << dst_reg << ", ";
    cout << (int)dst_reg+1 << "\n";
    cout << "before: " <<   
      m->core[reg_to_mem(dst_reg)] << "  " << 
      m->core[reg_to_mem(dst_reg+1)] << "\n";
#endif

    add_reg_pair (m, dst_reg, (data_val_t) dec->K);

#if 0
    cout << "after:  " <<   
      m->core[reg_to_mem(dst_reg)] << "  " << 
      m->core[reg_to_mem(dst_reg+1)] << "\n";
#endif

  } else {
    assert (dec->opcode == AVR_sbiw);
    sub_reg_pair (m, dst_reg, (data_val_t) dec->K);
  }
  
  int result_val =
    m->core[reg_to_mem(dst_reg)].val +
    m->core[reg_to_mem(dst_reg+(reg_t)1)].val << 8;
  int result_known =
    m->core[reg_to_mem(dst_reg)].known +
    m->core[reg_to_mem(dst_reg+(reg_t)1)].known << 8;
  
  bool n, n_known;

  if (result_known == 0xffff) {
    assign_status_bit (m, Z_BIT, result_val == 0);
    n = is_bit_set (result_val, 15);
    n_known = true;
    assign_status_bit (m, N_BIT, n);
  } else {
    kill_status_bit (m, Z_BIT);
    kill_status_bit (m, N_BIT);
    n = n_known = false;
  }
  
  bool v, v_known;

  if (is_bit_set (result_known, 15) && is_bit_set (orig_high_known, 7)) {
    bool c;
    if (dec->opcode == AVR_adiw) {
      c = is_bit_clear (result_val, 15) && is_bit_set (orig_high, 7);
      v = is_bit_set (result_val, 15) && is_bit_clear (orig_high, 7);
    } else {
      c = is_bit_set (result_val, 15) && is_bit_clear (orig_high, 7);
      v = is_bit_clear (result_val, 15) && is_bit_set (orig_high, 7);
    }
    v_known = true;
    assign_status_bit (m, C_BIT, c);
    assign_status_bit (m, V_BIT, v);
  } else {    
    kill_status_bit (m, C_BIT);
    kill_status_bit (m, V_BIT);
    v = v_known = false;
  }

  if (n_known && v_known) {
    bool s = n^v;
    assign_status_bit (m, S_BIT, s);
  } else {
    kill_status_bit (m, S_BIT);
  }

}

static inline void inc_x (memory *m) {
  inc_reg_pair (m, 26);
}

static inline void inc_y (memory *m) {
  inc_reg_pair (m, 28);
}

static inline void inc_z (memory *m) {
  inc_reg_pair (m, 30);
}

static inline void dec_x (memory *m) {
  dec_reg_pair (m, 26);
}

static inline void dec_y (memory *m) {
  dec_reg_pair (m, 28);
}

static inline void dec_z (memory *m) {
  dec_reg_pair (m, 30);
}

static void stack_dec (memory *m,
		       const struct decoder_info *dec,
		       int amt)
{
  if (debug_avr) printf ("decrementing SP by %d at %x\n", amt, dec->addr);
  m->shadow_sp -= amt;
#if TRACK_STACK
  for(int i=0; i<amt; ++i) {
    m->stack.push(mem_loc::bottom);
  }
#endif
}

static void stack_inc (memory *m,
		       const struct decoder_info *dec,
		       int amt)
{
  if (debug_avr) printf ("incrementing SP by %d at %x\n", amt, dec->addr);
  m->shadow_sp += amt;
#if TRACK_STACK
  for(int i=0; i<amt; ++i) {
    mem_loc x = m->stack.pop();
    // This should hold because inc/dec_SP should always be used in 
    // matched pairs.
    //assert (x == mem_loc::bottom);
  }
#endif
}

static int pair_to_num (data_val_t rv1, data_val_t rv2)
{
  return ((rv2 << 8) + rv1);
}

void abstract_interp_insn (memory *m,
			   const struct decoder_info *dec)
{
  /*
   * FIXME stupid hack: inefficient and sort of weird, but not sure
   * how else to do this: account for pushing of PC in first insn of
   * each interrupt handler
   */
  for (unsigned int i=0; i<platform->num_vectors; i++) {
    if (dec->addr == platform->int_vecs[i].start_addr) {
      if (platform->int_vecs[i].start_addr != (code_addr_t)0) {
	sub_mem_pair (m, io_to_mem (SPL_ADDR), PC_SIZE);
	m->shadow_sp -= PC_SIZE;
      }
    }
  }

  switch (dec->stack_opcode) {

  case AVR_DEC_STACK_PTR:
    switch (dec->stack_delta) {
    case DEC_UNK_1:
      {
	mem_loc_t r26 = m->core[reg_to_mem(26)];
	mem_loc_t r27 = m->core[reg_to_mem(27)];
	assert (r26.known == 0xff);
	assert (r27.known == 0xff);
	int amt = pair_to_num (r26.val, r27.val);
	stack_dec (m, dec, amt);
      }
      break;
    default:
      assert (dec->stack_delta > -4000 &&
	      dec->stack_delta < 4000);
      stack_dec (m, dec, dec->stack_delta);
      break;
    }
    break;

  case AVR_INC_STACK_PTR:
    switch (dec->stack_delta) {
    case INC_UNK_1:
      {
	mem_loc_t r28 = m->core[reg_to_mem(28)];
	mem_loc_t r29 = m->core[reg_to_mem(29)];
	mem_loc_t r30 = m->core[reg_to_mem(30)];
	mem_loc_t r01 = m->core[reg_to_mem( 1)];
	assert (r28.known == 0xff);
	assert (r29.known == 0xff);
	assert (r30.known == 0xff);
	assert (r01.known == 0xff);
	int amt1 = pair_to_num (r28.val, r29.val);
	int amt2 = pair_to_num (r30.val, r01.val);
	int new_sp = amt1 + amt2;
	int amt = new_sp - m->shadow_sp;
	stack_inc (m, dec, amt);
      }
      break;
    default:
      assert (dec->stack_delta > -4000 &&
	      dec->stack_delta < 4000);
      stack_inc (m, dec, dec->stack_delta);
      break;
    }
    break;

  case NO_STACK_OPCODE:
    break;

  default:
    die ();
  }

  switch (dec->opcode) {

  case AVR_SET_STACK_PTR:
#if TRACK_STACK
    m->stack.set_to_bottom();
#endif
    assert (dec->K_set);
    m->core[io_to_mem (SPH_ADDR)] = dec->K >> 8;
    m->core[io_to_mem (SPL_ADDR)] = dec->K & 0xff;
    m->stack_top = dec->K;
    m->shadow_sp = dec->K;
    break;

  case AVR_in:
    do_avr_in (m, dec);
    break;

  case AVR_out:
    do_avr_out (m, dec);
    break;

  case AVR_movw:
    do_avr_movw (m, dec);
    break;
    
#if !USE_BDD
  case AVR_mov:
    copy_mem_byte (m, 
		   reg_to_mem (dec->src_reg), 
		   reg_to_mem (dec->dst_reg));
    break;

  case AVR_ldi:
    assert (dec->K_set);
    set_val (m, 
	     reg_to_mem (dec->dst_reg),
	     (data_val_t) dec->K);
    break;
#endif

  case AVR_brcc: 
  case AVR_brcs: 
  case AVR_brne: 
  case AVR_breq: 
  case AVR_brpl: 
  case AVR_brmi: 
  case AVR_brvc: 
  case AVR_brvs: 
  case AVR_brge: 
  case AVR_brlt: 
  case AVR_brhc: 
  case AVR_brhs: 
  case AVR_brtc: 
  case AVR_brts: 
  case AVR_brid: 
  case AVR_brie: 
  case AVR_sbrs: 
  case AVR_sbrc: 
  case AVR_sbis: 
  case AVR_sbic: 
  case AVR_cpse:
    break;

    // these should have been covered
  case AVR_brbc:
  case AVR_brbs:
    die ();

    // these do not contribute to the dataflow analysis
  case AVR_jmp:
  case AVR_rjmp:
    assert(dec->target_set);
    break;

  case AVR_ijmp:
  case AVR_eijmp:
    assert (!dec->dst_reg_set);
    break;

  case AVR_rcall:
  case AVR_call:
    // FIXME: better to use stack model to push previous PC
    sub_mem_pair (m, io_to_mem (SPL_ADDR), PC_SIZE);
    m->shadow_sp -= PC_SIZE;
    break;

  case AVR_icall:
  case AVR_eicall:
    sub_mem_pair (m, io_to_mem (SPL_ADDR), PC_SIZE);
    m->shadow_sp -= PC_SIZE;
    assert (!dec->dst_reg_set);
    break;

  case AVR_reti:
    check_irqs_enabled(dec->addr, m);
    set_status_bit (m, I_BIT);
    // FALLTHROUGH!
  case AVR_ret:
    // FIXME: use stack model to pop the return addr and make sure
    // it's what we expect!
    add_mem_pair (m, io_to_mem (SPL_ADDR), PC_SIZE);
    m->shadow_sp += PC_SIZE;
    break;

  case AVR_adiw:
  case AVR_sbiw:
    arith_16 (m, dec);
    break;

#if USE_BDD
  case AVR_adc:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_adc (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_add:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_add (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_and:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_and (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_andi:
    assert (dec->dst_reg_set);
    assert (dec->K_set);
    bdd_bitbit_andi (m, dec->dst_reg, (data_val_t) dec->K);
    break;
  case AVR_asr:
    assert (dec->dst_reg_set);
    bdd_bitbit_asr (m, dec->dst_reg);
    break;
  case AVR_com:
    assert (dec->dst_reg_set);
    bdd_bitbit_com (m, dec->dst_reg);
    break;
  case AVR_cp:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_cp (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_cpc:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_cpc (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_cpi:
    assert (dec->dst_reg_set);
    assert (dec->K_set);
    bdd_bitbit_cpi (m, dec->dst_reg, (data_val_t) dec->K);
    break;
  case AVR_dec:
    assert (dec->dst_reg_set);
    bdd_bitbit_dec (m, dec->dst_reg);
    break;
  case AVR_eor:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_eor (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_inc:
    assert (dec->dst_reg_set);
    bdd_bitbit_inc (m, dec->dst_reg);
    break;
  case AVR_ldi:
    assert (dec->dst_reg_set);
    assert (dec->K_set);
    bdd_bitbit_ldi (m, dec->dst_reg, (data_val_t) dec->K);
    break;
  case AVR_lsr:
    assert (dec->dst_reg_set);
    bdd_bitbit_lsr (m, dec->dst_reg);
    break;
  case AVR_mov:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_mov (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_neg:
    assert (dec->dst_reg_set);
    bdd_bitbit_neg (m, dec->dst_reg);
    break;
  case AVR_or:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_or (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_ori:
    assert (dec->dst_reg_set);
    assert (dec->K_set);
    bdd_bitbit_ori (m, dec->dst_reg, (data_val_t) dec->K);
    break;
  case AVR_ror:
    assert (dec->dst_reg_set);
    bdd_bitbit_ror (m, dec->dst_reg);
    break;
  case AVR_sbc:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_sbc (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_sbci:
    assert (dec->dst_reg_set);
    assert (dec->K_set);
    bdd_bitbit_sbci (m, dec->dst_reg, (data_val_t) dec->K);
    break;
  case AVR_sub:
    assert (dec->dst_reg_set);
    assert (dec->src_reg_set);
    bdd_bitbit_sub (m, dec->dst_reg, dec->src_reg);
    break;
  case AVR_subi:
    assert (dec->dst_reg_set);
    assert (dec->K_set);
    bdd_bitbit_subi (m, dec->dst_reg, (data_val_t) dec->K);
    break;
  case AVR_swap:
    assert (dec->dst_reg_set);
    bdd_bitbit_swap (m, dec->dst_reg);
    break;

#endif

#if !USE_BDD
  case AVR_adc:
    // special case -- adc with same register is really rol -- rotate
    // left through carry
    assert (dec->src_reg_set && dec->dst_reg_set);
    if (dec->src_reg == dec->dst_reg) {
      logical_byte (m, dec);
    } else {
      arith_byte (m, dec);
    }
    break;

    // two register operands, operation only makes sense on whole bytes
  case AVR_add:
  case AVR_sbc:
  case AVR_sub:
    // one register, one immediate, operation only makes sense on whole bytes
  case AVR_subi:
  case AVR_sbci:
    // one operand
  case AVR_dec:
  case AVR_inc:
  case AVR_neg:
    arith_byte (m, dec);
    break;

    // one register, one immediate, bit-wise operation
  case AVR_ori:
  case AVR_andi:
    // two registers, bit-wise operation
  case AVR_and:
  case AVR_or:
  case AVR_eor:
    // one register, no immediate
  case AVR_swap:
  case AVR_com:
  case AVR_lsr:
  case AVR_asr:
  case AVR_ror:
    logical_byte (m, dec);
    break;
#endif

    /*
     * operations on condition code bits
     */

  case AVR_bld:
    assert (dec->b_set);
    assert (dec->dst_reg_set);
    copy_bit (m, 
	      io_to_mem (SREG_ADDR), 
	      6, 
	      reg_to_mem (dec->dst_reg), 
	      (data_t) dec->b);
    break;

  case AVR_bst:
    assert (dec->b_set);
    assert (dec->dst_reg_set);
    copy_bit (m, 
	      reg_to_mem (dec->dst_reg), 
	      (data_t) dec->b,
	      io_to_mem (SREG_ADDR), 
	      6);
    break;

  case AVR_clz:
    clear_status_bit (m, Z_BIT);
    break;

  case AVR_cln:
    clear_status_bit (m, N_BIT);
    break;

  case AVR_clv:
    clear_status_bit (m, V_BIT);
    break;

  case AVR_cls:
    clear_status_bit (m, S_BIT);
    break;

  case AVR_clh:
    clear_status_bit (m, H_BIT);
    break;

  case AVR_clt:
    clear_status_bit (m, T_BIT);
    break;

  case AVR_cli:
    clear_status_bit (m, I_BIT);
    break;

  case AVR_sec:
    set_status_bit (m, C_BIT);
    break;

  case AVR_sez:
    set_status_bit (m, Z_BIT);
    break;

  case AVR_sen:
    set_status_bit (m, N_BIT);
    break;

  case AVR_sev:
    set_status_bit (m, V_BIT);
    break;

  case AVR_ses:
    set_status_bit (m, S_BIT);
    break;

  case AVR_seh:
    set_status_bit (m, H_BIT);
    break;

  case AVR_set:
    set_status_bit (m, T_BIT);
    break;

  case AVR_sei:
    set_status_bit (m, I_BIT);
    break;

#if !USE_BDD
  case AVR_cp:
  case AVR_cpc:
  case AVR_cpi:
    do_avr_cp (m, dec);
    break;
#endif

    // all these put results in register pair 0,1
    // and kill C and Z in SREG
  case AVR_fmul:
  case AVR_fmuls:
  case AVR_fmulsu:
  case AVR_mul:
  case AVR_muls:
  case AVR_mulsu:
    // FIXME -- at least handle the case of concrete arguments
    kill_reg (m, (reg_t)0);
    kill_reg (m, (reg_t)1);
    kill_status_bit (m, C_BIT);
    kill_status_bit (m, Z_BIT);
    break;

  case AVR_cbi:
    assert (dec->ioaddr_set);
    assert (dec->b_set);
    clear_bit_addr (m, io_to_mem (dec->ioaddr), (data_t) dec->b);
    break;

  case AVR_sbi:
    assert (dec->ioaddr_set);
    assert (dec->b_set);
    set_bit_addr (m, io_to_mem (dec->ioaddr), (data_t) dec->b);    
    break;

  case AVR_sts:
    // not modeling memory yet
    break;

  case AVR_st1x:
    // X is unchanged
    break;
  case AVR_st2x:
    inc_x (m);
    break;
  case AVR_st3x:
    dec_x (m);
    break;

  case AVR_st1y:
  case AVR_std1y:
    // Y is unchanged
    break;
  case AVR_st2y:
    inc_y (m);
    break;
  case AVR_st3y:
    dec_y (m);
    break;

  case AVR_st1z:
  case AVR_std1z:
    // Z is unchanged
    break;
  case AVR_st2z:
    inc_z (m);
    break;
  case AVR_st3z:
    dec_z (m);
    break;

  case AVR_lds:
    // requires a memory model
    assert (dec->dst_reg_set);
    kill_reg (m, dec->dst_reg);
    break;

  case AVR_ld1x:
    kill_dest (m, dec);
    break;
  case AVR_ld2x:
    kill_dest (m, dec);
    inc_x (m);
    break;
  case AVR_ld3x:
    dec_x (m);
    kill_dest (m, dec);
    break;

  case AVR_ld1y:
  case AVR_ldd1y:
    kill_dest (m, dec);
    break;
  case AVR_ld2y:
    kill_dest (m, dec);
    inc_y (m);
    break;
  case AVR_ld3y:
    dec_y (m);
    kill_dest (m, dec);
    break;

  case AVR_ld1z:
  case AVR_ldd1z:
    kill_dest (m, dec);
    break;
  case AVR_ld2z:
    kill_dest (m, dec);
    inc_z (m);
    break;
  case AVR_ld3z:
    dec_z (m);
    kill_dest (m, dec);
    break;

  case AVR_pop:
    {
      if (debug_avr) printf ("pop at %x\n", dec->addr);
      inc_mem_pair (m, io_to_mem (SPL_ADDR));
      m->shadow_sp++;
#if TRACK_STACK
      mem_loc_t popped = m->stack.pop();
      m->core[reg_to_mem(dec->dst_reg)] = popped;
#else
      m->core[reg_to_mem(dec->dst_reg)] = mem_loc::bottom;
#endif
    }
    break;

  case AVR_push:
    if (debug_avr) printf ("push at %x\n", dec->addr);
    dec_mem_pair (m, io_to_mem (SPL_ADDR));
    m->shadow_sp--;
#if TRACK_STACK
    m->stack.push(m->core[reg_to_mem(dec->dst_reg)]);
#endif
    break;

  case AVR_elpm1:
  case AVR_lpm1:
    // FIXME -- need these to do jump tables
    kill_reg (m, (reg_t)0);
    break;

  case AVR_elpm2:
  case AVR_lpm2:
    // FIXME -- need these to do jump tables
    assert (dec->dst_reg_set);
    kill_reg (m, dec->dst_reg);
    break;

  case AVR_elpm3:
  case AVR_lpm3:
    // FIXME -- need these to do jump tables
    assert (dec->dst_reg_set);
    kill_reg (m, dec->dst_reg);
    kill_z (m);
    break;

    // these are pretty easy...
  case AVR_nop:
  case AVR_sleep:
  case AVR_wdr:
    break;

  case AVR_break:
  case AVR_spm:
    printf ("oops: illegal opcode %x at addr %x\n", dec->opcode, dec->addr);
    die ();
    
  default:
    printf ("oops: unknown opcode %x at addr %x\n", dec->opcode, dec->addr);
    die ();
  }
}


static void check_generic_branch (const state_inst *s, 
				  const struct control_info *c,
				  bool known_bit, 
				  bool taken,
				  code_addr_t addr)
{
  branch_idx_t index_taken = c->index_taken;
  branch_idx_t index_not_taken = c->index_not_taken;

  assert (index_taken != NO_IDX);
  assert (index_not_taken != NO_IDX);

  if (debug_avr) {
    unsigned int i;
    printf ("  check_generic_branch %x: taken= %d, not_taken= %d, len= %d, to: ",
	    addr, 
	    index_taken, 
	    index_not_taken, 
	    s->out_edges.len);
    for (i=0; i<s->out_edges.len; i++) {
      printf ("%x ", s->out_edges.list[i].target.addr);
    }
    printf ("\n");
  }

  assert (s->out_edges.len == 2);

  /*
   * if we know which way the branch is going, mark the other path as
   * dead
   */
  if (known_bit) {
    if (taken) {
      s->out_edges.list[index_taken].dead = false;	
      if (debug_avr) printf ("    taken edge isn't dead\n");
    } else {
      s->out_edges.list[index_not_taken].dead = false;	
      if (debug_avr) printf ("    not_taken edge isn't dead\n");
    }
  } else {
    s->out_edges.list[index_not_taken].dead = false;	
    s->out_edges.list[index_taken].dead = false;	
    if (debug_avr) printf ("    both edges are not dead\n");
  }
}

static bool bool_same (bool a, bool b)
{
  return !(a^b);
}

static void check_condition_code_branch (const state_inst *s, 
					 const struct control_info *c,
					 status_bit_t bit, 
					 bool taken,
					 code_addr_t addr)
{
  data_kno_t known = s->mem.core[io_to_mem(SREG_ADDR)].known;
  data_val_t val = s->mem.core[io_to_mem(SREG_ADDR)].val;
  bool bitset = is_bit_set (val, bit);

  check_generic_branch (s,
			c,
			is_bit_set (known, bit),
			bool_same (taken, bitset),
			addr);
}

void mark_dead_branches (const state_inst *s,
			 const struct decoder_info *dec,
			 const struct control_info *c)
{
  switch (dec->opcode) {

  case AVR_brcc:
    check_condition_code_branch (s, c, C_BIT, false, dec->addr);
    break;

  case AVR_brcs:
    check_condition_code_branch (s, c, C_BIT, true, dec->addr);
    break;

  case AVR_brne:
    check_condition_code_branch (s, c, Z_BIT, false, dec->addr);
    break;

  case AVR_breq:
    check_condition_code_branch (s, c, Z_BIT, true, dec->addr);
    break;

  case AVR_brpl:
    check_condition_code_branch (s, c, N_BIT, false, dec->addr);
    break;

  case AVR_brmi:
    check_condition_code_branch (s, c, N_BIT, true, dec->addr);
    break;

  case AVR_brvc:
    check_condition_code_branch (s, c, V_BIT, false, dec->addr);
    break;

  case AVR_brvs:
    check_condition_code_branch (s, c, V_BIT, true, dec->addr);
    break;

  case AVR_brge:
    check_condition_code_branch (s, c, S_BIT, false, dec->addr);
    break;

  case AVR_brlt:
    check_condition_code_branch (s, c, S_BIT, true, dec->addr);
    break;

  case AVR_brhc:
    check_condition_code_branch (s, c, H_BIT, false, dec->addr);
    break;

  case AVR_brhs:
    check_condition_code_branch (s, c, H_BIT, true, dec->addr);
    break;

  case AVR_brtc:
    check_condition_code_branch (s, c, T_BIT, false, dec->addr);
    break;

  case AVR_brts:
    check_condition_code_branch (s, c, T_BIT, true, dec->addr);
    break;

  case AVR_brid:
    check_condition_code_branch (s, c, I_BIT, false, dec->addr);
    break;

  case AVR_brie:
    check_condition_code_branch (s, c, I_BIT, true, dec->addr);
    break;

    // see branches above
  case AVR_sbrs:
  case AVR_sbrc:
  case AVR_sbis:
  case AVR_sbic:
    {
      bool taken, bitval, bitknown;
      data_val_t bit;
      assert (dec->b_set);
      bit = dec->b;
      
      if (dec->opcode == AVR_sbis ||
	  dec->opcode == AVR_sbic) {
	io_data_addr_t io;
	assert (dec->ioaddr_set);
	io = dec->ioaddr;
	bitval = is_bit_set (s->mem.core[io_to_mem(io)].val, bit);
	bitknown = is_bit_set (s->mem.core[io_to_mem(io)].known, bit);
      } else {
	// AVR_sbrs or AVR_sbrc
	reg_t reg;
	assert (dec->src_reg_set);
	reg = dec->src_reg;
	bitval = is_bit_set (s->mem.core[reg_to_mem(reg)].val, bit);
	bitknown = is_bit_set (s->mem.core[reg_to_mem(reg)].known, bit);
      }
      
      if (dec->opcode == AVR_sbrs ||
	  dec->opcode == AVR_sbis) {
	taken = bitval;
      } else {
	// AVR_sbrc or AVR_sbic
	taken = !bitval;
      }

      check_generic_branch (s, 
			    c,
			    bitknown,
			    taken,
			    dec->addr);
    }
    break;
   
  case AVR_cpse:
    {
      bool known, taken;
      reg_t src, dst;

      assert (dec->src_reg_set);
      assert (dec->dst_reg_set);

      src = dec->src_reg;
      dst = dec->dst_reg;
      
      known = 
	(s->mem.core[reg_to_mem(src)].known == 0xff) &&
	(s->mem.core[reg_to_mem(dst)].known == 0xff);

      taken = 
	s->mem.core[reg_to_mem(src)].val ==
	s->mem.core[reg_to_mem(dst)].val;
      
      check_generic_branch (s, c, known, taken, dec->addr);
    }
    break;

  case AVR_SET_STACK_PTR:
  case AVR_in:
  case AVR_out:
  case AVR_movw:
  case AVR_mov:
  case AVR_ldi:
  case AVR_jmp:
  case AVR_rjmp:
  case AVR_ijmp:
  case AVR_eijmp:
  case AVR_ret:
  case AVR_reti:
  case AVR_rcall:
  case AVR_call:
  case AVR_icall:
  case AVR_eicall:
  case AVR_adiw:
  case AVR_sbiw:
  case AVR_adc:
  case AVR_add:
  case AVR_sbc:
  case AVR_sub:
  case AVR_subi:
  case AVR_sbci:
  case AVR_dec:
  case AVR_inc:
  case AVR_neg:
  case AVR_ori:
  case AVR_andi:
  case AVR_and:
  case AVR_or:
  case AVR_eor:
  case AVR_swap:
  case AVR_com:
  case AVR_lsr:
  case AVR_asr:
  case AVR_ror:
  case AVR_bld:
  case AVR_bst:
  case AVR_clz:
  case AVR_cln:
  case AVR_clv:
  case AVR_cls:
  case AVR_clh:
  case AVR_clt:
  case AVR_cli:
  case AVR_sec:
  case AVR_sez:
  case AVR_sen:
  case AVR_sev:
  case AVR_ses:
  case AVR_seh:
  case AVR_set:
  case AVR_sei:
  case AVR_cp:
  case AVR_cpc:
  case AVR_cpi:
  case AVR_fmul:
  case AVR_fmuls:
  case AVR_fmulsu:
  case AVR_mul:
  case AVR_muls:
  case AVR_mulsu:
  case AVR_cbi:
  case AVR_sbi:

  case AVR_st1x:
  case AVR_st2x:
  case AVR_st3x:
  case AVR_st1y:
  case AVR_st2y:
  case AVR_st3y:
  case AVR_std1y:
  case AVR_st1z:
  case AVR_st2z:
  case AVR_st3z:
  case AVR_std1z:
  case AVR_sts:

  case AVR_ld1x:
  case AVR_ld2x:
  case AVR_ld3x:
  case AVR_ld1y:
  case AVR_ld2y:
  case AVR_ld3y:
  case AVR_ldd1y:
  case AVR_ld1z:
  case AVR_ld2z:
  case AVR_ld3z:
  case AVR_ldd1z:
  case AVR_lds:

  case AVR_pop:
  case AVR_push:

  case AVR_elpm1:
  case AVR_lpm1:
  case AVR_elpm2:
  case AVR_lpm2:
  case AVR_elpm3:
  case AVR_lpm3:
  case AVR_spm:

  case AVR_nop:
  case AVR_sleep:
  case AVR_wdr:
  case AVR_break:
    break;
    
  case AVR_brbc:
  case AVR_brbs:
    die();
    
  default:
    printf ("oops: unknown opcode %x at addr %x\n", dec->opcode, dec->addr);
    die ();
  }
}
