static bool calc_half_carry1 (data_val_t dst_val, 
			      data_val_t src_val, 
			      data_val_t result)
{
  return 
    (is_bit_set (dst_val, 3) && is_bit_set (src_val, 3)) ||
    (is_bit_set (src_val, 3) && !is_bit_set (result, 3)) ||
    (!is_bit_set (result, 3) && is_bit_set (dst_val, 3));
}

static bool calc_half_carry2 (data_val_t dst_val, 
			      data_val_t src_val, 
			      data_val_t result)
{
  return 
    (!is_bit_set (dst_val, 3) && is_bit_set (src_val, 3)) ||
    (is_bit_set (src_val, 3) && is_bit_set (result, 3)) ||
    (is_bit_set (result, 3) && !is_bit_set (dst_val, 3));
}

static bool calc_overflow1 (data_val_t dst_val, 
			    data_val_t src_val, 
			    data_val_t result)
{
  return
    (is_bit_set (dst_val, 7) && !is_bit_set (src_val, 7) && !is_bit_set (result, 7)) ||
    (!is_bit_set (dst_val, 7) && is_bit_set (src_val, 7) && is_bit_set (result, 7));
}

static bool calc_overflow2 (data_val_t dst_val, 
			    data_val_t src_val, 
			    data_val_t result)
{
  return
    (is_bit_set (dst_val, 7) && is_bit_set (src_val, 7) && !is_bit_set (result, 7)) ||
    (!is_bit_set (dst_val, 7) && !is_bit_set (src_val, 7) && is_bit_set (result, 7));
}

static bool calc_negative (data_val_t result)
{
  return is_bit_set (result, 7);
}

static bool calc_sign (bool n, bool v)
{
  return ((unsigned int)n ^ (unsigned int)v) != 0;
}

static bool calc_carry1 (data_val_t dst_val, 
			 data_val_t src_val, 
			 data_val_t result)
{
  return 
    (((dst_val & 0x80) == 0) && ((src_val & 0x80) != 0)) ||
    (((src_val & 0x80) != 0) && ((result & 0x80) != 0)) ||
    (((result & 0x80) != 0) && ((dst_val & 0x80) == 0));	  
}

static bool calc_carry2 (data_val_t dst_val, 
			 data_val_t src_val, 
			 data_val_t result)
{
  return 
    (((dst_val & 0x80) != 0) && ((src_val & 0x80) != 0)) ||
    (((src_val & 0x80) != 0) && ((result & 0x80) == 0)) ||
    (((result & 0x80) == 0) && ((dst_val & 0x80) != 0));	  
}

static void set_z_bit (struct memory *m, 
		       data_kno_t res_known, 
		       data_val_t res_val)
{
  // Z bit
  if (res_known == 0xff) {
    if (res_val == 0) {
      set_status_bit (m, Z_BIT);
    } else {
      clear_status_bit (m, Z_BIT);
    }
  } else {
    kill_status_bit (m, Z_BIT);
  }
}

static void split_nibble (data_t in,
			  data_t *low,
			  data_t *high)
{
  *low = in & 0xf;
  *high = (in >> 4) & 0xf;
}

static void logical_byte (struct memory *m,
			  const struct decoder_info *dec)
{
  reg_t dst_reg;
  data_val_t dst_val;
  reg_t src_reg = NO_REG;
  data_val_t src_val = 0;
  bool src_val_set = false;
  data_kno_t src_known = 0;
  data_val_t res_val;
  data_kno_t res_known, dst_known;
      
  assert (dec->dst_reg_set);
  dst_reg = dec->dst_reg;
  dst_val = m->core[reg_to_mem(dst_reg)].val;
  dst_known = m->core[reg_to_mem(dst_reg)].known;

  switch (dec->opcode) {

  case AVR_ori:
  case AVR_andi:
    assert (dec->K_set);
    src_val = (data_val_t) dec->K;
    src_val_set = true;
    src_known = 0xff;
    break;

  case AVR_or:
  case AVR_and:
  case AVR_eor:
    assert (dec->src_reg_set);
    src_reg = dec->src_reg;
    src_val = m->core[reg_to_mem(src_reg)].val;
    src_val_set = true;
    src_known = m->core[reg_to_mem(src_reg)].known;
    break;

  case AVR_lsr:
  case AVR_asr:
  case AVR_ror:
  case AVR_adc: // special case where adc == rol
  case AVR_com:
  case AVR_swap:
    break;

  default:
    die ();
  }

  switch (dec->opcode) {

  case AVR_adc: // special case where adc means rol -- rotate left
		// through carry

    res_val = dst_val << 1;
    res_known = dst_known << 1;

    if (is_status_bit_known (m, C_BIT)) {
      res_known |= 0x1;
      if (get_status_bit (m, C_BIT)) {
	res_val |= 0x1;
      } else {
	res_val &= 0x1fe;
      }
    } else {
      res_known &= 0x1fe;
    }
    if ((res_known & 0x100) != 0) {
      if ((res_val & 0x100) != 0) {
	set_status_bit (m, C_BIT);
      } else {
	clear_status_bit (m, C_BIT);
      }
    } else {
      kill_status_bit (m, C_BIT);
    }
    res_val &= 0xff;
    res_known &= 0xff;
    break;

  case AVR_ror:
  case AVR_lsr:
  case AVR_asr:
    res_val = 0x0;
    res_known = 0x0;
    break;

  case AVR_com:
    res_val = 0xff & (~dst_val);
    res_known = dst_known;
    break;

  case AVR_swap:
    {
      data_val_t low, high;
      data_kno_t low_known, high_known;
      
      split_nibble (dst_val, &low, &high);
      split_nibble (dst_known, &low_known, &high_known);
      
      res_val = (low << 4) & high;
      res_known = (low_known << 4) & high_known;
    }
    break;

  case AVR_eor:
    assert (src_val_set);
    // special case -- handle this idiom for zeroing a register
    if (src_reg == dst_reg) {
      res_val = 0x0;
      res_known = 0xff;
    } else {
      res_known = src_known & dst_known;
      res_val = res_known & (dst_val ^ src_val);
    }
    break;

  case AVR_ori:
  case AVR_or:
    assert (src_val_set);
    if (src_reg == dst_reg) {
      res_val = src_val;
      res_known = src_known;
    } else {
      res_val = dst_val | src_val;
      res_known = 
	(src_val & src_known) | 
	(dst_val & dst_known) | 
	(~src_val & src_known & ~dst_val & dst_known);
    }
    break;

  case AVR_andi:
  case AVR_and:
    assert (src_val_set);
    if (src_reg == dst_reg) {
      res_val = src_val;
      res_known = src_known;
    } else {
      res_val = dst_val & src_val;
      res_known = 
	(~src_val & src_known) | 
	(~dst_val & dst_known) | 
	(src_val & src_known & dst_val & dst_known);
    }
    break;

  default:
    die ();
  }

  assert (res_val == (res_val & 0xff));
  assert (res_known == (res_known & 0xff));
  assert (dst_reg != NO_REG);

  m->core[reg_to_mem(dst_reg)].val = res_val & res_known;
  m->core[reg_to_mem(dst_reg)].known = res_known;

  // sets no condition codes
  if (dec->opcode == AVR_swap) return;
  
  // not like the others...
  if (dec->opcode == AVR_adc) {
    set_z_bit (m, res_known, res_val);

    kill_status_bit (m, N_BIT);
    kill_status_bit (m, V_BIT);
    kill_status_bit (m, S_BIT);
    kill_status_bit (m, H_BIT);

    return;
  }

  kill_status_bit (m, C_BIT); 
  kill_status_bit (m, V_BIT);  // zero the V bit

  // N and S bits, where N==S since V==0
  if ((res_known & 0x80) != 0) {
    if ((res_val & 0x80) != 0) {
      set_status_bit (m, N_BIT);
      set_status_bit (m, S_BIT);
    } else {
      clear_status_bit (m, N_BIT);
      clear_status_bit (m, S_BIT);
    }
  } else {
    kill_status_bit (m, N_BIT);
    kill_status_bit (m, S_BIT);
  }

  set_z_bit (m, res_known, res_val);
}

static void do_avr_cp (struct memory *m,
		       const struct decoder_info *dec)
{
  reg_t src = dec->src_reg;
  reg_t dst = dec->dst_reg;

  assert (dst < 32);
  
  data_val_t src_val, dst_val, result;
  bool s, v, n, z, c, h;
  bool prev_carry, prev_zero, guess_carry;
  
  assert (dec->src_reg_set || dec->K_set);
  assert (dec->dst_reg_set);
  
  if ((dec->opcode != AVR_cpi && (m->core[reg_to_mem(src)].known != 0xff)) ||
      (m->core[reg_to_mem(dst)].known != 0xff) ||
      ((dec->opcode == AVR_cpc) &&
       (((m->core[io_to_mem(SREG_ADDR)].known & 0x1) == 0) ||
	((m->core[io_to_mem(SREG_ADDR)].known & 0x2) == 0))
       )) {
    kill_status_bit (m, C_BIT);
    kill_status_bit (m, Z_BIT);
    kill_status_bit (m, N_BIT);
    kill_status_bit (m, V_BIT);
    kill_status_bit (m, S_BIT);
    kill_status_bit (m, H_BIT);

    return;
  }
  
  if (dec->opcode == AVR_cpi) {
    assert (dec->K_set);
    src_val = (data_val_t) dec->K;
  } else {
    src_val = m->core[reg_to_mem(src)].val;
  }
  dst_val = m->core[reg_to_mem(dst)].val;

  prev_carry = ((m->core[io_to_mem(SREG_ADDR)].val & 0x1) != 0); 
  prev_zero = ((m->core[io_to_mem(SREG_ADDR)].val & 0x2) != 0);
  
  result = dst_val - src_val;
  if (dec->opcode == AVR_cpc) {
    result -= (prev_carry?1:0);
  }
  guess_carry = ((result & 0x100) != 0);
  result &= 0xff;
  
  h = calc_half_carry2 (dst_val, src_val, result);
  v = calc_overflow1 (dst_val, src_val, result);
  n = calc_negative (result);
  s = calc_sign (n, v);
  if (dec->opcode == AVR_cpc) {
    z = (result == 0) ? prev_zero : false;
  } else {
    z = result == 0;
  }
  c = calc_carry1 (dst_val, src_val, result);
  assert ((c && guess_carry) || (!c && !guess_carry));
  
  assign_status_bit (m, C_BIT, c);
  assign_status_bit (m, Z_BIT, z);
  assign_status_bit (m, N_BIT, n);
  assign_status_bit (m, V_BIT, v);
  assign_status_bit (m, S_BIT, s);
  assign_status_bit (m, H_BIT, h);
}


static void arith_byte (struct memory *m,
			const struct decoder_info *dec)
{
  reg_t dst_reg;
  data_val_t src_val = 0;
  data_val_t dst_val;
  data_val_t result;
  bool carry, guess_carry;
  bool src_val_set = false;

  assert (dec->dst_reg_set);
  dst_reg = dec->dst_reg;

  if (m->core[reg_to_mem(dst_reg)].known != 0xff) goto bail;
  dst_val = m->core[reg_to_mem(dst_reg)].val;

  if (dec->opcode == AVR_sbci ||
      dec->opcode == AVR_adc ||
      dec->opcode == AVR_sbc) {
    if ((m->core[io_to_mem(SREG_ADDR)].known & 0x1) == 0) goto bail;
    carry = ((m->core[io_to_mem(SREG_ADDR)].val & 0x1) != 0);
  } else {
    carry = false;
  }

  switch (dec->opcode) {

  case AVR_subi:
  case AVR_sbci:
    assert (dec->K_set);
    src_val = (data_val_t) dec->K;
    src_val_set = true;
    break;

  case AVR_inc: 
  case AVR_dec:
    break;

  case AVR_neg:
    src_val = 0x0;
    src_val_set = true;
    break;

  case AVR_adc:
  case AVR_add:
  case AVR_sbc:
  case AVR_sub:
    {
      reg_t src_reg;
      assert (dec->src_reg_set);
      src_reg = dec->src_reg;
      if (m->core[reg_to_mem(src_reg)].known != 0xff) goto bail;	  
      src_val = m->core[reg_to_mem(src_reg)].val;	
      src_val_set = true;
    }
    break;

  default:
    printf ("oops: can't handle %s\n", opcode_to_string (dec->opcode));
    die ();
  }
	
  switch (dec->opcode) {

  case AVR_neg:
    result = (~dst_val) + 1;  // two's complement negation
    break;

  case AVR_adc:
    assert (src_val_set);
    result = src_val + dst_val + ((carry)?1:0);
    break;

  case AVR_dec:
    result = dst_val - 1;
    break;

  case AVR_inc:
    result = dst_val + 1;
    break;

  case AVR_add:
    assert (src_val_set);
    result = src_val + dst_val;
    break;

  case AVR_sbc:
  case AVR_sbci:
    assert (src_val_set);
    result = dst_val - (src_val + ((carry)?1:0));
    break;

  case AVR_sub:
  case AVR_subi:
    assert (src_val_set);
    result = dst_val - src_val;
    break;

  default:
    die ();
  }

  guess_carry = ((result & 0x100) != 0);

  if (debug_avr) printf ("  src = %x, dst = %x, result = %x, guess_carry = %s\n",
			 src_val, dst_val, result, bool_to_str(guess_carry));

  result &= 0xff;
  
  m->core[reg_to_mem(dst_reg)].val = result;

  if (dec->opcode == AVR_inc ||
      dec->opcode == AVR_dec) {

    bool z, n, v, s;

    z = result == 0;
    n = calc_negative (result);

    if (dec->opcode == AVR_dec) {
      if (result == 0x7f) {
	v = true;
	} else {
	  v = false;
	}
    } else {
      if (result == 0x80) {
	v = true;
      } else {
	v = false;
      }
    }
      
    s = calc_sign (n, v);

    assign_status_bit (m, Z_BIT, z);
    assign_status_bit (m, N_BIT, n);
    assign_status_bit (m, V_BIT, v);
    assign_status_bit (m, S_BIT, s);

  } else {
    bool h, v, n, s, z, c;
        
    assert (src_val_set);

    switch (dec->opcode) {
    case AVR_neg:
      c = !(result == 0);
      v = (result == 0x80);
      h = is_bit_set (dst_val, 3) || is_bit_set (result, 3);
      break;
      
    case AVR_adc:
    case AVR_add:
      c = calc_carry2 (dst_val, src_val, result);
      v = calc_overflow2 (dst_val, src_val, result);
      h = calc_half_carry1 (dst_val, src_val, result);
      break;
      
    case AVR_sbc:
    case AVR_sbci:
    case AVR_sub:
    case AVR_subi:
      c = calc_carry1 (dst_val, src_val, result);
      v = calc_overflow1 (dst_val, src_val, result);
      h = calc_half_carry2 (dst_val, src_val, result);
      break;
      
    default:
      die ();
    }

    n = calc_negative (result);
    s = calc_sign (n, v);
    z = result == 0;

    if (dec->opcode == AVR_sbc ||
	dec->opcode == AVR_sbci) {

      if (is_status_bit_known (m, Z_BIT)) {
	if (z && get_status_bit (m, Z_BIT)) {
	  set_status_bit (m, Z_BIT);
	} else {
	  clear_status_bit (m, Z_BIT);
	}
      } else {
	kill_status_bit (m, Z_BIT);
      }
    } else {
      assign_status_bit (m, Z_BIT, z);
    }

    assign_status_bit (m, C_BIT, c);
    assign_status_bit (m, N_BIT, n);
    assign_status_bit (m, V_BIT, v);
    assign_status_bit (m, S_BIT, s);
    assign_status_bit (m, H_BIT, h);
  }
  return;
  
 bail:      
  kill_reg (m, dst_reg);
  kill_status_bit (m, C_BIT);
  kill_status_bit (m, Z_BIT);
  kill_status_bit (m, N_BIT);
  kill_status_bit (m, V_BIT);
  kill_status_bit (m, S_BIT);
  kill_status_bit (m, H_BIT);
}

