#include "stacktool.h"

#include <fcntl.h>
#include <unistd.h>
#include <libelf/libelf.h>

#if 1

// On Windows, we may not have stab.h so we have a copy here.  Blech!

// All these stab types have string pointers in their n_value field 
// so we have to watch out for them
#define N_GSYM  0x20
#define N_FUN   0x24
#define N_STSYM 0x26
#define N_LCSYM 0x28
#define N_OPT   0x3c
#define N_RSYM  0x40
#define N_SLINE 0x44
#define N_SO    0x64
#define N_LSYM  0x80
#define N_BINCL 0x82
#define N_SOL   0x84
#define N_PSYM  0xa0
#define N_EINCL 0xa2
#define N_LBRAC 0xc0
#define N_RBRAC 0xe0

#else

// Since we are cross-compiling so it is reasonable to ask if this
// file is the right one to look at.  On my Linux box, the only
// difference between /usr/include/stab.h (and related includes) and
// the one that ships with avr-gcc are typos in the comments and
// tweaks to the license.
#include <stab.h>

#endif

static void elf_die (void) __attribute__ ((noreturn));

static unsigned char *raw_text = NULL;
unsigned int code_size = (unsigned int)-1;

queue<srcloc> *srclocs = 0;  // size == code_size

static Elf32_Sym *symtab = NULL;
static unsigned int symtab_len = (unsigned int)-1;

#define NO_SEC ((Elf32_Half)-1)

static char *strtab = NULL;
static Elf32_Half text_sec_num = NO_SEC;

size_t data_seg_size = (size_t)0;
size_t bss_seg_size = (size_t)0;

// According to bfd/stabs.c, stabs entries for 32-bit object formats
// are 12-byte objects with the following format.
// ToDo: But avr isn't a 32-bit processor.
//
// The stabs symbols are divided into compilation units.  For the
// first entry in each unit, the type of 0, the value is the length of
// the string table for this unit, and the desc field is the number of
// stabs symbols for this unit.
typedef struct {
        Elf32_Word    n_name;  // stab string table index (see sh_link)
        unsigned char n_type;
        unsigned char n_other;
        Elf32_Half    n_desc;
        Elf32_Word    n_value;
} Stab;

// This is based on code from knit/compiler/rename_dot_o_files.c for
// renaming stab entries.  It should provide some kind of starting
// point.

static char* elf_get_strings(Elf *elf, int i)
{
  Elf32_Shdr *shdr;
  Elf_Scn  *scn; // need this to interpret strings
  Elf_Data *data;

  scn = elf_getscn(elf, i);
  shdr = elf32_getshdr(scn);
  if (shdr->sh_type != SHT_STRTAB) {
    cerr << "No string table" << endl;
    elf_die();
  }
  data = 0;
  if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
    cerr << "Missing/empty string table" << endl;
    elf_die();
  }
  return (char*)data->d_buf;
}

void load_stabs(Elf *elf, Elf_Scn* scn)
{
  Elf32_Shdr *shdr;
  Elf_Data *data;

  int last_table = -1;

  if (scn == 0) return; // That was easy...

  shdr = elf32_getshdr(scn);
#if 0 // not one of the recognised formats
  if (shdr->sh_type != SHT_SYMTAB) {
    panic("Not a symbol table");
  }
#endif
  char* strings = elf_get_strings(elf,shdr->sh_link);

  data = 0;
  if ((data = elf_getdata(scn, data)) == 0 || data->d_size == 0) {
    cerr << "Missing/empty symbol table" << endl;
    elf_die();
  }

  Stab *stabs = (Stab*)data->d_buf;
  int n_stabs = data->d_size / sizeof(Stab);

  code_addr_t last_addr = (code_addr_t) code_size;
  char* cur_filename = 0;
  int cur_line = -1;

  int fun_start = 0;  // start of current function
  int offset = 0;
  int next_offset = 0;
  for( int i = 0; i < n_stabs; ++i ) {
    switch (stabs[i].n_type) {
    case 0:
      {
        offset += next_offset;
        next_offset = stabs[i].n_value;
        // printf("Num stab symbols: %d\n", stabs[i].n_desc);
#if 1 // more to do here
        if (last_table != -1) {
          cerr << "Multi-part (partially linked?) .o files can't be handled yet"
               << endl;
          elf_die();
        }
#endif
        last_table = i;
        goto next;
      }
    case N_SLINE :
      {
        code_addr_t addr = fun_start + stabs[i].n_value;
        assert(srclocs);
        assert( addr < (code_addr_t) code_size );
        if (cur_filename && cur_line != -1) {
          for(code_addr_t j = last_addr; j<addr; ++j) {
            srclocs[j].insert(srcloc(cur_filename,cur_line));
          }
        }
        last_addr = addr+1;
        cur_line = stabs[i].n_desc;
        srclocs[addr].insert(srcloc(cur_filename,cur_line));
        if (debug_elf) {
          printf("Line number: %d starts at 0x%x\n", stabs[i].n_desc, addr );
        }
      }
      break;
    case N_SOL   :
      {
        code_addr_t addr = stabs[i].n_value;
        assert(srclocs);
        assert( addr < (code_addr_t) code_size );
        if (cur_filename && cur_line != -1) {
          for(code_addr_t j = last_addr; j<addr; ++j) {
            srclocs[j].insert(srcloc(cur_filename,cur_line));
          }
        }
        last_addr = (code_addr_t) code_size;
        cur_filename = drop_srcloc_prefix(strings + offset + stabs[i].n_name);
        if (debug_elf) {
          printf("Filename: '%s' starts at 0x%x\n", cur_filename, addr);
        }
      }
      break;
    case N_BINCL   :
      {
        // ToDo: does this drop some srclocs on the floor?
        cur_filename = drop_srcloc_prefix(strings + offset + stabs[i].n_name);
        if (debug_elf) {
          printf("Filename: '%s'\n", cur_filename);
        }
      }
      break;
    case N_FUN:
      cur_line = stabs[i].n_desc;
      if (stabs[i].n_desc == 0) { // last instruction of function
        code_addr_t addr = stabs[i].n_value;
        if (cur_filename && cur_line != -1) {
          for(code_addr_t j = last_addr; j<=addr; ++j) {
            srclocs[j].insert(srcloc(cur_filename,cur_line));
          }
        }
        fun_start = 0;
        last_addr = (code_addr_t) code_size;
      } else {
        last_addr = fun_start = stabs[i].n_value;
      }
      break;
    case N_SO    :
    case N_GSYM:
    case N_STSYM:
    case N_LCSYM:
    case N_OPT   :
    case N_LSYM  :
    case N_RSYM  :
    case N_PSYM  :
      //    case N_BINCL :
    case N_EINCL :
    case N_LBRAC :
    case N_RBRAC :
      {
#if 0
        printf("Stab: 0x%x '%s'\n", stabs[i].n_type, strings + offset + stabs[i].n_name);
#endif
      }
      break;
    default:
      {
        //        printf("Unrecognised stab type: 0x%x '%s'\n", stabs[i].n_type, strings + offset + stabs[i].n_name);                        
      }
      break;
    }
  next:;
  }
}
        
unsigned short get_insn_short (code_addr_t addr)
{
  unsigned short ret = 
    raw_text[addr] |
    (raw_text[addr+1] << 8);
  return ret;
}

bool addr_to_srclocs (code_addr_t addr, queue<srcloc> **locs)
{
  if (addr == NO_CODE_ADDR 
      || addr >= (code_addr_t) code_size
      || srclocs[addr].empty()) {
    locs = 0;
    return false;
  } else {
    assert(srclocs);
    *locs = &srclocs[addr];
    return true;
  }
}

bool addr_to_srcloc (code_addr_t addr, srcloc *loc)
{
  if (addr == NO_CODE_ADDR 
      || addr >= (code_addr_t) code_size) {
    loc->filename = 0;
    loc->line_number = -1;
    return false;
  } else {
    assert(srclocs);
    return srclocs[addr].last(loc);
  }
}

ostream& pp_addr (ostream& os, code_addr_t addr) {
  srcloc loc;
  if (addr_to_srcloc(addr,&loc)) {
    return os << loc << " (" << addr << ")";
  } else {
    return os << addr;
  }
};

code_addr_t find_first_srcloc (const srcloc &loc) {
  for(code_addr_t i=0; i<code_size; ++i) {
    if (srclocs[i].contains(loc)) {
      return i;
    }
  }
  return NO_CODE_ADDR;
}
      
code_addr_t find_last_srcloc (const srcloc &loc) {
  for(code_addr_t i=code_size-1; (int)i>=0; --i) {
    if (srclocs[i].contains(loc)) {
      return prevPC(i+1);
    }
  }
  return NO_CODE_ADDR;
}

static const char * sh_type_name (Elf32_Word sht)
{
  switch (sht) {
  case 0: return "SHT_NULL";
  case 1: return "SHT_PROGBITS";
  case 2: return "SHT_SYMTAB";
  case 3: return "SHT_STRTAB";	
  case 4: return "SHT_RELA";
  case 5: return "SHT_HASH";	
  case 6: return "SHT_DYNAMIC";		
  case 7: return "SHT_NOTE";
  case 8: return "SHT_NOBITS";		
  case 9: return "SHT_REL";	
  case 10: return "SHT_SHLIB";		
  case 11: return "SHT_DYNSYM";	
  case 14: return "SHT_INIT_ARRAY";	
  case 15: return "SHT_FINI_ARRAY";	
  case 16: return "SHT_PREINIT_ARRAY";
  case 17: return "SHT_GROUP";
  case 18: return "SHT_SYMTAB_SHNDX";
  case 19: return "SHT_NUM";
  default:
    assert (false);
  }
  return "";
}	

void open_obj_file (const char *fn)
{
  Elf32_Ehdr *ehdr;
  Elf *elf;
  Elf_Scn *scn;
  Elf_Data *data;
  int fd;
  Elf32_Half cnt;

  /* Open the input file */
  fd = open (fn, O_RDONLY);
  if (fd == -1) {
    printf ("oops: couldn't open '%s'\n", fn);
    die ();
  }

  /* Obtain the ELF descriptor */
  (void) elf_version (EV_CURRENT);
  elf = elf_begin (fd, ELF_C_READ, NULL);
  if (elf == NULL) {
    elf_die ();
  }

  /* Obtain the .shstrtab data buffer */
  ehdr = elf32_getehdr (elf);
  if (ehdr == NULL) elf_die ();

  if (debug_elf) {
    printf ("ehdr->e_machine = %d\n", ehdr->e_machine);
  }

  // make sure we're talking about an AVR here
  // assert (ehdr->e_machine == EM_AVR);

  scn = elf_getscn (elf, (size_t) ehdr->e_shstrndx);
  if (scn == NULL) elf_die ();

  data = elf_getdata (scn, NULL);
  if (data == NULL) elf_die ();

  if (debug_elf) {
    printf ("sections:\n");
  }

  cnt = (Elf32_Half) 0;
  scn = NULL; 

  while ((scn = elf_nextscn (elf, scn)) != NULL) {
    char *name;
    Elf_Data *scn_data;
    Elf32_Shdr *shdr;
    void *dbuf;
    Elf_Type dtype;
    size_t dsize;

    cnt++;

    shdr = elf32_getshdr (scn);
    if (shdr == NULL) {
      elf_die();
    }

    name = (char *) data->d_buf + shdr->sh_name;
    scn_data = elf_getdata (scn, NULL);
    if (scn_data == NULL) elf_die ();

    dbuf = scn_data->d_buf;
    dtype = scn_data->d_type;
    dsize = scn_data->d_size;

    if (debug_elf) {
      printf ("\n");
      printf ("  [%d]\n", cnt);    
      printf ("    sh_name:      %s\n", name);
      printf ("    sh_type:      %s\n", sh_type_name (shdr->sh_type));
      printf ("    sh_flags:     %x\n", shdr->sh_flags);
      printf ("    sh_addr:      %x\n", shdr->sh_addr);
      printf ("    sh_offset:    %x\n", shdr->sh_offset);
      printf ("    sh_size:      %x\n", shdr->sh_size);
      printf ("    sh_link:      %x\n", shdr->sh_link);
      printf ("    sh_info:      %x\n", shdr->sh_info);
      printf ("    sh_addralign: %x\n", shdr->sh_addralign);
      printf ("    sh_entsize:   %x\n", shdr->sh_entsize);
      printf ("\n");
      printf ("    d_buf:  %p\n", dbuf);
      printf ("    d_type: %x\n", dtype);
      printf ("    d_size: %x\n", dsize);
    }

    if (strcmp (name, ".data") == 0) {
      data_seg_size = dsize;
    }

    if (strcmp (name, ".bss") == 0) {
      bss_seg_size = dsize;
    }

    if (strcmp (name, ".text") == 0) {
      text_sec_num = cnt;
      raw_text = (unsigned char *) dbuf;
      code_size = dsize;
      srclocs = new queue<srcloc>[code_size];
      if (debug_elf) {
	printf ("    code_size = %x\n", code_size);
      }
    }

    if (strcmp (name, ".symtab") == 0) {
      symtab = (Elf32_Sym *) dbuf;
      symtab_len = dsize / sizeof (Elf32_Sym);
      if (debug_elf) {
	printf ("    symtab_len = %x\n", symtab_len);
      }
    }

    if (strcmp (name, ".strtab") == 0) {
      strtab = (char *) dbuf;
    }

    if (strcmp (name, ".stab") == 0) {
      load_stabs(elf,scn);
    }

  }

  assert (raw_text != NULL);
  assert (strtab != NULL);
  assert (symtab != NULL);

  close (fd);
}

static char *lookup_symbol (Elf32_Half sec, code_addr_t addr)
{
  unsigned int i;
  char *ret = NULL;
  
  for (i=0; i<symtab_len; i++) {
    char *name = strtab + symtab[i].st_name;
    Elf32_Half section = symtab[i].st_shndx;

    assert (name != NULL);

    if ((section == sec) &&
	(symtab[i].st_value == (Elf32_Addr) addr) &&
	(strlen (name) > 0)) {
      ret = name;
    }
  }
  
  return ret;
}

char *lookup_text_symbol (code_addr_t addr)
{
  assert (text_sec_num != NO_SEC);
  return lookup_symbol (text_sec_num, addr);
}

code_addr_t label_to_addr (const char* label) {
  assert (text_sec_num != NO_SEC);
  Elf32_Half sec = text_sec_num;

  // cout << endl;

  for (unsigned int i=0; i<symtab_len; i++) {
    char *name = strtab + symtab[i].st_name;
    Elf32_Half section = symtab[i].st_shndx;
    Elf32_Addr addr    = symtab[i].st_value;
    assert (name != NULL);

    // cout << "'" << name << "' '" << label << "'" << endl;

    if ((section == sec) &&
	(0 == strcmp(name,label))) {
      return (code_addr_t)addr;
    }
  }

  // cout << endl;

  return NO_CODE_ADDR;
}

static void elf_die (void)
{
  fprintf (stderr, "%s\n", elf_errmsg (elf_errno()));
  die ();
}
