////////////////////////////////////////////////////////////////
// List which can be accessed in FIFO and LIFO order
////////////////////////////////////////////////////////////////

typedef unsigned int slot_t;

static slot_t inc_slot(slot_t x, slot_t max) {
  x++;
  if (x==max) {
    return 0;
  } else {
    return x;
  }
};

static slot_t dec_slot(slot_t x, slot_t max) {
  if (x==0) {
    return max-1;
  } else {
    return x-1;
  }
};

template<class Element> class queue_iterator;

template<class Element>
class queue {

 private:

  Element *q;
  slot_t size;
  slot_t head; // first element in circular queue
  slot_t tail; // last element+1
  
  void grow(void) { // increase size of queue
    slot_t new_size = 2*size+1; // any strictly increasing function will do here
    // cout << "Growing queue from " << dec << size << " to " << new_size << hex << endl;
    Element *nq = new Element[new_size];
    slot_t i;
    slot_t j;
    for(i = 0, j = head; j != tail; ++i, j=inc_slot(j,size)) {
      assert(i < new_size && j < size);
      nq[i] = q[j];
    }
    delete[] q;
    size = new_size;
    head = 0;
    tail = i;
    q = nq;
  };
     

 public:

  typedef queue_iterator<Element> iterator;

  queue<Element>(slot_t _size=1) : size(_size), head(0), tail(0) {
    assert(size>0); // inc_slot fails if size==0
    q = new Element[size];
  };

  bool empty() const { return (head==tail); };

  int length() const { 
    return (head <= tail) ? (tail-head) : (size-head+tail);      
  };

  void push (const Element &e) {
    slot_t ix = inc_slot(tail,size);
    if ( ix == head ) {
      this->grow();
      ix = inc_slot(tail,size);
    }
    assert( ix != head );
    q[tail] = e;
    tail = ix;
  };

  bool pop (Element *e) {
    if (this->empty()) {
      return false;
    } else {
      tail = dec_slot(tail,size);
      *e = q[tail];
      return true;
    }
  };

  bool first (Element *e) const {
    if (this->empty()) {
      return false;
    } else {
      *e = q[head];
      return true;
    }
  };

  bool last (Element *e) const {
    if (this->empty()) {
      return false;
    } else {
      *e = q[dec_slot(tail,size)];
      return true;
    }
  };

  bool contains (const Element &e) const {
    for(slot_t i = head; i != tail; i = inc_slot(i,size)) {
      if (e == q[i]) {
        return true;
      }
    }
    return false;
  };

  bool insert (const Element &e) {
    bool changed = !this->contains(e);
    if (changed) this->push(e);
    return changed;
  };

  void unshift (const Element &e) {
    slot_t ix = dec_slot(head,size);
    if ( ix == tail ) {
      this->grow();
      ix = dec_slot(head,size);
    }
    assert( ix != tail );
    head = ix;
    q[head] = e;
  };

  bool shift (Element *e) {
    if (this->empty()) {
      return false;
    } else {
      *e = q[head];
      head = inc_slot(head,size);
      return true;
    }
  };

  iterator begin() const;
  iterator end()   const;
  friend iterator iterator::operator ++();
  friend Element queue_iterator<Element>::operator*();

  bool operator <=(const class queue<Element> &y) const {
    for(class queue<Element>::iterator i = this->begin(); i != this->end(); ++i) {
      if (!y.contains(*i)) return false;
    }
    return true;
  };

};

////////////////////////////////////////////////////////////////
// Iterator
////////////////////////////////////////////////////////////////

template<class Element> 
class queue_iterator {
  slot_t x;
  const queue<Element> &s;
 public:
  queue_iterator<Element>(const queue<Element> &_s, slot_t _x) 
    : x(_x), s(_s) 
    {};
  Element operator*() {
    assert(x != s.tail);
    return s.q[x];
  };
  queue_iterator<Element> operator ++ () {
    x = inc_slot(x,s.size);
    return *this;
  };
  bool operator == (const class queue_iterator<Element> b) const {
    return (x == b.x);
  };
  bool operator != (const class queue_iterator<Element> b) const {
    return !(*this == b);
  };
};

template<class Element> 
inline queue_iterator<Element> queue<Element>::begin() const {
  return queue_iterator<Element>(*this,this->head);
};

template<class Element> 
inline queue_iterator<Element> queue<Element>::end() const {
  return queue_iterator<Element>(*this,this->tail);
};

////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////

template<class Element>
ostream& pp_list (ostream& os, const class queue<Element> &b, string sep) {
  for(class queue<Element>::iterator i = b.begin();
      i != b.end();
      ++i) {
    os << *i << sep;
  }
  return os;
};

template<class Element>
ostream& operator << (ostream& os, const class queue<Element> &b) {
  return pp_list(os, b, " ");
};

// returns false if all elements were already in dst
// (We don't call it union because that is a keyword....)
template<class Element>
inline
bool union2(class queue<Element> &dst, const class queue<Element> &src) {
  bool changed = false;
  for(class queue<Element>::iterator x = src.begin(); x != src.end(); ++x) {
    changed |= dst.insert(*x);
  }
  return changed;
};

template<class Element>
bool disjoint(const class queue<Element> &x, const class queue<Element> &y) {
  for(class queue<Element>::iterator i = y.begin(); i != y.end(); ++i) {
    if (x.contains(*i)) return false;
  }
  return true;
};

////////////////////////////////////////////////////////////////
// End
////////////////////////////////////////////////////////////////
