////////////////////////////////////////////////////////////////
// Abstract Values
//
// We do all intermediate calculations with 32-bit values 
// and then extract the bits we need at the end.
//
// As long as all abstract values are built using 'set_val'
// (I think they are), we maintain the invariant that 
// all unknown bits are zero.
// Note that there is no representation for 'top' - the handling
// of control flow in the rest of the abstract interpreter avoids
// the need to manipulate such values.
////////////////////////////////////////////////////////////////

#include "rtl_ops.h"

// moved to header file
// typedef struct {
//   int known;  // bitmask
//   int val;    // value
// } val;

typedef unsigned int uint;

void rtl_copy( val *dst, val *src )
{
  rtl_set( dst, src->known, src->val );
}

// form greatest lower bound
void rtl_merge( val *dst, val *src1, val *src2 )
{
  rtl_set( dst, 
           src1->known & src2->known & ~(src1->val ^ src2->val),
           src1->val | src2->val );
}

void rtl_bottom( val *dst )
{
  rtl_set( dst, 0, 0);
}

void rtl_set( val *dst, int known, int v ) 
{
  dst->known = known;
  dst->val   = v & known;
}

int rtl_val( val *x )
{
  return (x->val & x->known);
}

// fill unknown bits with 1 to get maximum (unsigned) concrete value
int rtl_maxUVal( val *x )
{
  return ((x->val & x->known) | ~x->known);
}

// maximum (signed) concrete value
int rtl_maxVal( val *x )
{
  int sign = 0x80000000;
  if (x->known & sign) {
    return ((x->val & x->known) | ~x->known);
  } else {
    return (((x->val & x->known) | ~x->known) & ~sign);
  }
}

// minimum (signed) concrete value
int rtl_minVal( val *x )
{
  int sign = 0x80000000;
  if (x->known & sign) {
    return (x->val & x->known);
  } else {
    return ((x->val & x->known) | 0x80000000);
  }
}

void rtl_const( val *dst, int v) {
  rtl_set(dst,~0,v);
}

////////////////////////////////////////////////////////////////
// Sign- and zero- extension
////////////////////////////////////////////////////////////////

// zero extend an n-bit value
void rtl_zx(val *dst, val *src, int n)
{
  int mask = (1<<n)-1;
  rtl_set(dst, (src->known & mask) | ~mask, src->val & mask);
}

// sign extend an n-bit value
void rtl_sx(val *dst, val *src, int n)
{
  int mask = (1 << n) - 1;
  int sign = 1 << (n-1);
  rtl_set(dst, 
          ((src->known & mask) | ((src->known & sign) ? ~mask : 0)),
          ((src->val   & mask) | ((src->val   & sign) ? ~mask : 0)));
}

// extract n bits starting with bit s
void rtl_extract_bits(val *dst, val *src, int s, int n) {
  int mask = (1<<n)-1;
  rtl_set(dst,
          ((src->known >> s) & mask) | ~mask,
          (src->val   >> s) & mask);
}

// extract 1 bit
void rtl_get_bit(val *dst, val *src, int b) {
  rtl_extract_bits(dst, src, b, 1);
}

// set 1 bit
void rtl_put_bit(val *dst, val *src1, int b, val *src2) {
  int mask = 1<<b;
  rtl_set(dst,
          (src1->known & ~mask) | ((src2->known << b) & mask),
          (src1->val   & ~mask) | ((src2->val   << b) & mask));
}

// bottom 8 bits of value
int rtl_val8( val *x )
{
  return (x->val & x->known & 0xff);
}

////////////////////////////////////////////////////////////////
// Arithmetic operations
////////////////////////////////////////////////////////////////

static void rtl_arith1(val *dst, val *src, int r);
static void rtl_arith2(val *dst, val *src1, val *src2, int r);

// single-operand arithmetic operations set their result using
// this function.
// Note that we get a 9-bit result from an 8-bit operation
static void rtl_arith1(val *dst, val *src, int r)
{
  // ToDo: we can get accurate results on the bottom N bits if
  // the bottom N bits of the argument are set
  if ((src->known & 0xffff)==0xffff) {
    rtl_set(dst,0x1ffff,r);
  } else if ((src->known & 0xff)==0xff) {
    rtl_set(dst,0x1ff,r);
  } else {
    rtl_bottom(dst);
  }
}

// two-operand arithmetic operations set their result using
// this function.
static void rtl_arith2(val *dst, val *src1, val *src2, int r)
{
  if ((src1->known & src2->known & 0xffff) == 0xffff) {
    rtl_set(dst,0x1ffff,r);
  } else if ((src1->known & src2->known & 0xff) == 0xff) {
    rtl_set(dst,0x1ff,r);
  } else {
    rtl_bottom(dst);
  }
}

void rtl_neg(val *dst, val *src) { 
  rtl_arith1(dst, src, -src->val); 
}

void rtl_add(val *dst, val *src1, val *src2) {
  //  printf("add: (%0x,%0x)",src1->known,src2->known);
  rtl_arith2(dst, src1, src2, src1->val+src2->val); 
  // printf(" = %0x\n",dst->known);
}

void rtl_sub(val *dst, val *src1, val *src2) {
  rtl_arith2(dst, src1, src2, src1->val-src2->val); 
}

void rtl_div(val *dst, val *src1, val *src2) {
  rtl_bottom(dst);
  //  rtl_arith2(dst, src1, src2, src1->val / src2->val); 
}

void rtl_mod(val *dst, val *src1, val *src2) {
  rtl_bottom(dst);
  //  rtl_arith2(dst, src1, src2, src1->val % src2->val); 
}

void rtl_mul(val *dst, val *src1, val *src2) {
  rtl_bottom(dst);
  //  rtl_arith2(dst, src1, src2, src1->val * src2->val); 
}

void rtl_mulu(val *dst, val *src1, val *src2) {
  rtl_bottom(dst);
  //  rtl_arith2(dst, src1, src2, (uint)((uint)src1->val * (uint)src2->val)); 
}

////////////////////////////////////////////////////////////////
// Logical operations
////////////////////////////////////////////////////////////////

void rtl_com(val *dst, val *src) {
  rtl_set(dst,src->known,~src->val);
}

void rtl_and(val *dst, val *src1, val *src2) {
  rtl_set(dst, 
          ( (~src1->val & src1->known)
          | (~src2->val & src2->known)
          | (src1->val & src1->known & src2->val & src2->known)),
          src1->val & src2->val);
}

void rtl_or(val *dst, val *src1, val *src2) {
  rtl_set(dst, 
          ( (src1->val & src1->known)
          | (src2->val & src2->known)
          | (~src1->val & src1->known & ~src2->val & src2->known)),
          src1->val | src2->val);
}

void rtl_xor(val *dst, val *src1, val *src2) {
  rtl_set(dst, src1->known & src2->known, src1->val ^ src2->val);
}

void rtl_shra(val *dst, int n, val *src1, val *src2) {
  rtl_bottom(dst);
}

void rtl_shrl(val *dst, int n, val *src1, val *src2) {
  rtl_bottom(dst);
}

void rtl_rotr(val *dst, val *src1, val *src2) {
  rtl_bottom(dst);
}

void rtl_rolc(val *dst, val *src1, val *src2) {
  rtl_bottom(dst);
}

////////////////////////////////////////////////////////////////
// Comparisions
////////////////////////////////////////////////////////////////

// ToDo: Could give a definitive "no" if ranges don't
// overlap (or if known bits disagree)
void rtl_eq(val *dst, int n, val *src1, val *src2) {
  int mask = (1 << n) - 1;
  if ((src1->known & src2->known & mask) == mask) {
    rtl_const(dst,((src1->val & mask) == (src2->val & mask)) ? 1 : 0);
  } else {
    rtl_bottom(dst);
  }
}

void rtl_lt(val *dst, int n, val *src1, val *src2) {
  val x1, x2;
  rtl_sx(&x1,src1,n);
  rtl_sx(&x2,src2,n);
  if (rtl_maxVal(&x1) < rtl_minVal(&x2)) {
    rtl_const(dst,1);
  } else if (rtl_minVal(&x1) >= rtl_maxVal(&x2)) {
    rtl_const(dst,0);
  } else {
    rtl_bottom(dst);
  }
}

void rtl_le(val *dst, int n, val *src1, val *src2) {
  val x1, x2;
  rtl_sx(&x1,src1,n);
  rtl_sx(&x2,src2,n);
  if (rtl_maxVal(&x1) <= rtl_minVal(&x2)) {
    rtl_const(dst,1);
  } else if (rtl_minVal(&x1) > rtl_maxVal(&x2)) {
    rtl_const(dst,0);
  } else {
    rtl_bottom(dst);
  }
}

void rtl_gt(val *dst, int n, val *src1, val *src2) {
  rtl_lt(dst,n,src2,src1);
}

void rtl_ge(val *dst, int n, val *src1, val *src2) {
  rtl_le(dst,n,src2,src1);
}

////////////////////////////////////////////////////////////////
// Conditionals
////////////////////////////////////////////////////////////////

void rtl_cond(val *dst, val *cond, val *src1, val *src2) {
  if (cond->known & 1) {
    rtl_copy(dst, (cond->val & 1) ? src1 : src2 );
  } else {
    rtl_merge(dst, src1, src2);
  }
}

////////////////////////////////////////////////////////////////
// End
////////////////////////////////////////////////////////////////

