#include "stacktool.h"

int bdd_op_inputs[32];
queue<call_node> call_queue;

/*
 * everything important happens here
 */
static void flow_analysis (void)
{
  int pass = 0, csteps, dsteps;
  int total_steps = 0;

  control_flow_setup ();

  /*
   * keep doing control-flow and data-flow analysis until an overall
   * fixpoint is reached 
   */
  while (true) {
    pass++;

    csteps = 0;
    while (true) {
      bool cont = control_flow_step ();
      if (!cont) break;
      csteps++;
    }

    // Hack: this should be outside the loop but we don't want to run
    // it until after we have found an icall.  Blech!
    if (pass==1) add_tinyos_tasks ();

    dsteps = 0;
    while (true) {
      bool cont = dataflow_step ();
      if (!cont) break;
      dsteps++;
    }
    
    total_steps += dsteps;

    // This is so that calls are handled AFTER a local fixed point is reached
    call_node_t element;
    while (call_queue.pop(&element)){
      deferred_add_tinyos_task (&element);
    }	

    if (csteps == 0 && dsteps == 0) break;

    check_dataflow_invariants ();

#if VERBOSE_CTRL_FLOW
    if (!quiet) printf ("analysis pass %d:  %d control flow steps, %d abstract interp steps\n",
			pass, csteps, dsteps);
#endif

  }

#if 0
  // There are still some false positives for this check
  if(check_inlined_task_posts()==false){ 
      cout << "ERROR: Bad task table. Terminating!\n" << endl;
      die();
  }
#endif

  if (!quiet) {
    cout << dec << total_steps << hex << " abstract interp steps" << endl;
  }
  if (print_callgraph) {
    callgraph_shapes ();
  }
  
  report_dead_branches ();

#if VERBOSE_CTRL_FLOW
  report_stores ();
#endif
  
  // whole-program analysis
  stack_scan ();
#if WCET_ANALYSIS
  if (do_wcet) {
    wcet_analysis ();
  }
#endif
}

static void print_int (const struct interrupt_vector *iv)
{
  printf ("  name = \"%s\"\n", iv->name);
  printf ("  start_addr = %x\n", iv->start_addr);
  printf ("  enable: addr = %x, bit = %x\n", iv->enable.addr, iv->enable.bit);
  printf ("  flag:   addr = %x, bit = %x\n", iv->flag.addr, iv->flag.bit);
}

static bool int_eq (const struct interrupt_vector *i1,
		    const struct interrupt_vector *i2)
{
  if (
      ((i1->name == NULL && i2->name == NULL) ||
       (strcmp (i1->name, i2->name) == 0)) &&
      (i1->start_addr == i2->start_addr) &&
      (i1->enable.addr == i2->enable.addr) &&
      (i1->enable.bit == i2->enable.bit) &&
      (i1->flag.addr == i2->flag.addr) &&
      (i1->flag.bit == i2->flag.bit)
      ) {
    return true;
  } else {
    return false;
  }
}

int main (const int argc, 
	  const char *argv[])
{
  const char *input_file;
  cout << hex;

  {
    // has to be done before parsing args
    int x;
    for (x=0; x<MAX_VECTORS; x++) {
      int_info[x].max_reenter = 1;
    }
  }

  input_file = parse_args (argc, argv);

  if (platform == NULL) {
    printf ("oops: need to specify a platform using -mmcu=xxx\n");
    exit (-1);
  }

  // sanity check
  {
    struct interrupt_vector no_int = NO_INT;
    unsigned int i;
    for (i=0; i<platform->num_vectors; i++) {

      if (verbose && debug_control > 0) {
	printf ("i=%d\n", i);
	print_int (&platform->int_vecs[i]);
      }

      assert (!int_eq (&no_int, &platform->int_vecs[i]));
      // true on currently supported AVRs, but not on some others
      assert (platform->int_vecs[i].start_addr == (code_addr_t)(i * VECTOR_SIZE));
    }

    if (verbose && debug_control > 0) {
      printf ("i=%d\n", platform->num_vectors);
      print_int (&platform->int_vecs[platform->num_vectors]);
    }
    assert (int_eq (&platform->int_vecs[platform->num_vectors],
		    &no_int));
  }

  SRAM_TOP = platform->sram.offset + (data_addr_t) (platform->sram.len - 1);

  open_obj_file (input_file);  
  open_aux_files (input_file);
  flow_analysis ();

  if (print_disasm) {
    if (disasm_fn) {
      disassemble (disasm_fn);
    } else {
      char* fn = strdup (input_file);
      strcpy (fn+strlen(fn)-4, ".out");
      disassemble (fn);
      free (fn);
    }
  }

  if (!quiet) find_counts ();

  check_schedulers();

  close_aux_files ();

  if (!quiet) {
    printf ("\n*** normal termination ***\n");
    fflush (stderr);
    fflush (stdout);
  }

  return 0;
}
