#include "stacktool.h"

// all special code to handle tinyos goes here

// it's basically all about the task scheduler!


// Note that there are two parts to tasks.
// 1) We keep track of which task schedulers we know of in schedulers[]
// 2) We keep track of potential icall targets in tos_tasks
//    (A static inside add_icall_target)
//
// indirect function calls that are not TinyOS tasks are treated 
// completely separately

class scheduler schedulers[MAX_TASKS];
int num_schedulers = 0;

class scheduler* find_scheduler(code_addr_t addr)
{
  for(int i=0; i<num_schedulers; ++i) {
    if (schedulers[i].addr == addr) {
      return &schedulers[i];
    }
  }
  return NULL;
}

static const char* known_schedulers[] = {
  (const char*)"TOSH_run_next_task",
  (const char*)"TOSH_run_task",
  (const char*)"file1_TOSH_run_next_task",
  (const char*)"file1_TOSH_run_task",
  (const char*)"TOS_schedule_task",
  (const char*)"main",  // inlining moves TOS scheduler to here
  NULL
};

bool is_scheduler (code_addr_t addr)
{
  // printf ("is_scheduler (%x)   ", addr);

  // FIXME hack! -- assume we're in main if there's no address
  if (addr == NO_ADDR || addr == 0) {
    addr = label_to_addr ("main");
    assert (addr);
  }

  // printf ("is_scheduler (%x)\n", addr);

  char *fn = lookup_text_symbol (addr);
  int i=0;
  while (known_schedulers[i]) {
    if (strcmp (known_schedulers[i], fn) == 0) return true;
    i++;
  }
  return false;
}

void check_schedulers (void)
{
  /*
   * TinyOS-specific hack.  To handle tasks correctly, we need to
   * be able to find task addresses. 
   */
  int unrecognized = 0;
  for(int i=0; i<num_schedulers; ++i) {
    code_addr_t last_labelled = schedulers[i].addr;

    // find label
    while(last_labelled>0 && 
          (all_code[last_labelled]==0 || all_code[last_labelled]->decode.label == NULL)) {
      --last_labelled;
    }
    const char* label = 
      (last_labelled == 0 && all_code[last_labelled])
      ? "" 
      : all_code[last_labelled]->decode.label;
    
    // check if it is in the list of known callers
    for(int j=0; known_schedulers[j]; ++j) {
      if (0 == strcmp(label, known_schedulers[j])) {
        goto known;
      }
    }
    printf("WARNING WARNING WARNING WARNING WARNING WARNING WARNING\n");
    printf("Unrecognized icall found.\n");
    printf("  scheduler: 0x%04x (== %s+%d)\n", 
             schedulers[i].addr, 
             label, schedulers[i].addr-last_labelled);
    printf("Try to add a test to control.c:check_schedulers() to eliminate this warning in the future.\n");
    printf("WARNING WARNING WARNING WARNING WARNING WARNING WARNING\n");
    unrecognized++;
  known:;
  }  

  if (unrecognized > 0) die();
}

void add_scheduler (code_addr_t addr, 
		    abs_loc_t ret, 
		    abs_loc_t head)
{
  if (debug_control > 0) {
    cout << "Adding scheduler " << hex << addr << " : " << head << "/" << ret << endl;
  }
  if (find_scheduler(addr)) {
    return;
  }
  schedulers[num_schedulers].addr = addr;
  // icallers[num_icallers].func = head.addr;
  schedulers[num_schedulers].stash_ret  = ret;
  schedulers[num_schedulers].stash_head = head;
  num_schedulers++;
}

set<code_addr_t,MAX_TASKS> tos_tasks;

// FIXME: why is this called from two places?

// This is for keeping track of potential tasks.  The
// schedulers[] data structure keeps track of potential sources of
// icalls (tinyos schedulers).
bool add_tinyos_task (code_addr_t task_addr)
{
  // hack -- always call tasks from context 0 -- this is valid for TinyOS

  // avoid duplication
  if (tos_tasks.insert(task_addr)) {

    if (debug_control) {
      cout << "found an icall to 0x" << task_addr << " " << lookup_text_symbol (task_addr) << endl;
      printf ("first time seeing icall to %x\n", 
	      task_addr);
    }

    enqueue_task (task_addr);
    return true;

  } else {
    return false;
  }
}

queue<char*> icall_funs;

void add_tinyos_tasks(void) {
  for(queue<char*>::iterator i = icall_funs.begin(); i != icall_funs.end(); ++i) {
    code_addr_t fn = label_to_addr(*i);
    if (fn == NO_CODE_ADDR) {
      cout << "WARNING: Bogus icallee '" << (*i) << "' ignored" << endl;
    } else {
      add_tinyos_task (fn);
    }
  }
}

// Table of higher order functions used to help handle icalls
static struct ho_fn_t ho_functions[] = {
  { (char*)"TOS_post",                     (reg_t)24, },
  { (char*)"tinyos_task_scheduler_1_post", (reg_t)24, },
  { 0, 0 }
};

struct ho_fn_t* lookup_ho_fn (code_addr_t adr)
{
  // FIXME hack! -- assume we're in main if there's no address
  if (adr == NO_ADDR) {
    adr = label_to_addr ("main");
    assert (adr);
  }
  
  assert (all_code[adr] != NULL);
  for (int i=0; ho_functions[i].name != 0; i++) {
    if (all_code[adr] &&
	all_code[adr]->decode.label &&
	strcmp (all_code[adr]->decode.label, 
		(char*)(ho_functions[i].name)) == 0) {
      return &ho_functions[i];
    }
  }
  return NULL;
}

void deferred_add_tinyos_task (call_node_t *call)
{
  static bool printed_warning = false;

  if (debug_data > 0) {
    printf ("handling call to (apparent) higher-order fn %x\n", call->called_addr);
  }

  struct ho_fn_t *hofn = lookup_ho_fn (call->called_addr);
  if (!hofn) {
    die();
  }
  
  char *fun = hofn->name;
  reg_t reg = hofn->arg;
  
  mem_loc_t rlo = *regster(call->calling_pc, reg);
  mem_loc_t rhi = *regster(call->calling_pc, reg+1);

  if (rlo.known != 0xff || rhi.known != 0xff) {    
    // suppress duplicates
    if (!printed_warning) {
      printf ("**************************************************************\n");
      printf ("OOPS -- found call (from %x) to TinyOS task '%s' with\n",
	      call->calling_pc.addr,
	      fun);
      printf ("inconclusive arguments (%x:%x %x:%x)!\n",
	      rlo.known, rlo.val,
	      rhi.known, rhi.val);
      printf ("\n");
      printf ("This happens sometimes because stacktool lacks a path-sensitive\n");
      printf ("analysis.  It means that the analysis may be missing tasks and\n");
      printf ("therefore unsound.  You should create a file called x.cnt (where you\n");
      printf ("pick x) and fill it with lines like \"ICall TimerM$HandleFire\", one\n");
      printf ("for each task.  Then put x.cnt on the stacktool command line.\n");
      printf ("**************************************************************\n");
      printed_warning = true;
    }
    return;
  }

  code_addr_t task_addr = pair_to_addr (rlo.val, rhi.val);

  if (add_tinyos_task (task_addr)) {
      printf ("adding tinyos task data for post %x (%s), task %x (%s)\n",
	      call->called_addr, lookup_text_symbol (call->called_addr),
	      task_addr, lookup_text_symbol (task_addr));
  }

}

